/*
 * Copyright (C) 1995  Sam Moore, Warren Jasper
 * Copyright (C) 1997,1998,1999,2000,2001,2002,2003,2004  Tim Edwards
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

/*----------------------------------------------------------------*/
/* ciodio96.c							  */
/*----------------------------------------------------------------*/

#include <linux/autoconf.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/timer.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/signal.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/uaccess.h>

/* Check for PCI in kernel compilation */

#if defined(CONFIG_PCI)
#define ENABLE_PCI

/* Support for ICS Advent boards 4/28/04 */
#ifndef PCI_VENDOR_ID_ICS_ADVENT
#define PCI_VENDOR_ID_ICS_ADVENT	0x494F
#endif

#ifndef PCI_VENDOR_ID_COMPUTERBOARDS
#define PCI_VENDOR_ID_COMPUTERBOARDS	0x1307
#endif

#ifndef PCI_DEVICE_ID_PCIDIO96
#define PCI_DEVICE_ID_PCIDIO96 		0x0054
#endif

#ifndef PCI_DEVICE_ID_PCIDIO48H
#define PCI_DEVICE_ID_PCIDIO48H		0x000B
#endif

#ifndef PCI_DEVICE_ID_PCIDIO96H
#define PCI_DEVICE_ID_PCIDIO96H		0x0017
#endif

#ifndef PCI_DEVICE_ID_PCIDIO24
#define PCI_DEVICE_ID_PCIDIO24 		0x0028
#endif

#ifndef PCI_DEVICE_ID_PCIDIO24H
#define PCI_DEVICE_ID_PCIDIO24H		0x0014
#endif

/* ICS Advent boards */
#ifndef PCI_DEVICE_ID_PCIDIO24DB
#define PCI_DEVICE_ID_PCIDIO24DB	0x0C51
#endif

#ifndef PCI_DEVICE_ID_PCIDIO24DB_CT
#define PCI_DEVICE_ID_PCIDIO24DB_CT	0x0E52
#endif

#ifndef PCI_DEVICE_ID_PCIDIO24HD
#define PCI_DEVICE_ID_PCIDIO24HD	0x0C50
#endif

#ifndef PCI_DEVICE_ID_PCIDIO24HD_CT
#define PCI_DEVICE_ID_PCIDIO24HD_CT	0x0E51
#endif

#ifdef ADAPTER_ID
#undef ADAPTER_ID
#define ADAPTER_ID "PCI_DIO"
#endif

#include <linux/pci.h>

#endif

#include "dio96.h"

#ifdef get_fs_long
#define get_long(x,addr) (x) = get_fs_long(addr)
#else
#define get_long get_user
#endif

#ifdef put_fs_long
#define put_long put_fs_long
#else
#define put_long put_user
#endif

/***************************************************************************
 *
 * Prototype of pubic and private functions.
 *
 ***************************************************************************/

#define BYTE unsigned char
#define LONG unsigned long

static ssize_t dio_read(struct file *filePtr, char *buf, size_t count, loff_t *ppos);
static ssize_t dio_write(struct file *filePtr, const char *buf, size_t count, loff_t *ppos);

static int dio_open(struct inode *iNode, struct file *filePtr);
static int dio_close(struct inode *iNode, struct file *filePtr);
static int dio_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg);

int init_module(void);
void cleanup_module(void) ;

/***************************************************************************
 *
 * Global data.
 *
 ***************************************************************************/

/* Default device major number */

#ifndef DEFAULT_MAJOR_DEV
#define DEFAULT_MAJOR_DEV  31    /* Default Major Device Number */
#endif

/* Base I/O Address for ISA-bus cards */

#ifndef BASE_REG
#define BASE_REG         0x300   /* Default base address for CIO-DIO cards */
#endif

/* Number of channels */

#ifndef CHANNELS
#define CHANNELS	6
#endif

#define A_OFFSET	0
#define B_OFFSET	1
#define C_OFFSET	2
#define CTRL_OFFSET     3
#define PORT_OFFSET	4

#define A_MASK		0xff
#define B_MASK		0xff00
#define C_MASK		0xff0000

typedef struct CHANNEL_REC {
   int           open;        /* for lockout of other processes */
   unsigned char mode;	      /* for readback of 82C55 mode */
   unsigned char bits[3];     /* saved state of output */
} ChanRec;

static unsigned char channels;
static unsigned int dio_base_reg;
static unsigned char dio_has_bit_ops;	    /* Bit-wise operations enabled  */
static ChanRec Chan[CHANNELS];              /* Channel specific information */
static int MajorNumber = DEFAULT_MAJOR_DEV; /* Major number compiled in     */

static spinlock_t driver_lock = SPIN_LOCK_UNLOCKED;

#ifndef TRUE
#define TRUE  1
#endif
#ifndef FALSE
#define FALSE 0
#endif

/***************************************************************************
 *
 * read() service function
 *
 ***************************************************************************/

static ssize_t dio_read(struct file *filePtr, char *buf, size_t count, loff_t *ppos)
{
  int   minor = MINOR(filePtr->f_dentry->d_inode->i_rdev);

  long  value = 0L;
  int   portbase = dio_base_reg + (minor * PORT_OFFSET);

  if (minor < channels) {
    if (count > 0) value = inb(portbase + A_OFFSET);
    if (count > 1) value += (inb(portbase + B_OFFSET) << 8);
    if (count > 2) value += (inb(portbase + C_OFFSET) << 16);

	spin_lock_irq(&driver_lock);
    put_long(value, buf);
    spin_unlock_irq(&driver_lock);
    return count;
  }

  printk("dio_read: Incorrect minor number (%d).\n", minor);
  return -1;
}

/***************************************************************************
 *
 * write() service function
 *
 ***************************************************************************/

static ssize_t dio_write(struct file *filePtr, const char *buf, size_t count,
	loff_t *ppos)
{
    int  minor = MINOR(filePtr->f_dentry->d_inode->i_rdev);

    int  portbase = dio_base_reg + (minor * PORT_OFFSET);
    long value;
    BYTE bpart;

    if ( minor >= channels ) {
        printk("dio_write: bad minor number = %d\n", minor );
        return -1;
    }
    spin_lock_irq(&driver_lock);
    get_long(value, buf);
    value &= 0xffffff;
    bpart = (BYTE)(value & A_MASK);
    Chan[minor].bits[0] = bpart;
    outb_p( bpart,        portbase + A_OFFSET);
    bpart = (BYTE)((value & B_MASK) >> 8);
    Chan[minor].bits[1] = bpart;
    outb_p( bpart,  portbase + B_OFFSET);
    bpart = (BYTE)((value & C_MASK) >> 16);
    Chan[minor].bits[2] = bpart;
    outb_p( bpart, portbase + C_OFFSET);
    spin_unlock_irq(&driver_lock);
    #ifdef DEBUG
        printk("DIO set to %#x\n", value & 0xf);
    #endif
    return 3;
}


/***************************************************************************
 *
 * open() service handler
 *
 ***************************************************************************/

static int dio_open(struct inode *iNode, struct file *filePtr)
{
    int minor = MINOR(iNode->i_rdev);
    int portbase = dio_base_reg + (minor * PORT_OFFSET);

    /* 
       check if device is already open: only one process may read from a
       port at a time.  There is still the possibility of two processes 
       reading from two different channels messing things up. However,
       the overhead to check for this may not be worth it.
    */

    if ( Chan[minor].open == TRUE ) {
        return -EBUSY;
    }

    Chan[minor].open = TRUE;              /* The device is open */
    Chan[minor].mode = 0;

#ifdef HAVE_TIMER
    if (minor == 5) {
       outb_p(0x0, portbase);
       outb_p(0x0, portbase + 1);
    }
    else
       outb_p(0x0, portbase + CTRL_OFFSET);
#else
    outb_p(0x0, portbase + CTRL_OFFSET);
#endif

    #ifdef DEBUG
        printk(ADAPTER_ID ": open(): minor %d mode %d.\n", minor, Chan[minor].mode);
    #endif

    return (0);   
}

/***************************************************************************
 *
 * close() service handler
 *
 ***************************************************************************/

static int dio_close(struct inode *iNode, struct file *filePtr)
{
    int minor = MINOR(iNode->i_rdev);

    #ifdef DEBUG
        printk(ADAPTER_ID ": close() of minor number %d.\n", minor);
    #endif

    Chan[minor].open = FALSE;

    return (0);
}

/***************************************************************************
 *
 * iotctl() service handler
 *
 ***************************************************************************/

/* Note:
    Remember that FIOCLEX, FIONCLEX, FIONBIO, and FIOASYN are reserved ioctl cmd numbers
*/

/*------------------------------------------------------------------------------*/
/* IMPORTANT:  In this revision, ioctl() IGNORES the "minor" value and operates */
/* on the correct control register based on the ioctl command, provided that	*/
/* device has been opened (must be opened so it can be initialized. . .)	*/
/*										*/
/* ioctl() does NOT check to see if port has been opened already.		*/
/* ioctl() takes advantage of the 82C55's own bit-level commands for port C	*/
/*										*/
/* bit-wise and mode operations should not be allowed on the counter and	*/
/* interrupt registers of the 8253/8254, but for reasons of speed this is not	*/
/* checked.  Illegal registers can be written to, so it is up to the programmer */
/* to avoid the problem.							*/ 
/*------------------------------------------------------------------------------*/

static int dio_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd,
	LONG arg)
{
  int minor = MINOR(iNode->i_rdev);
  int portbase = dio_base_reg + (minor << 2);  /* PORT_OFFSET hard-coded */
  int channel = (cmd & 0x018) >> 3;

  /*--------------------------------------------------------------------*/
  /* ioctl command setup:						*/
  /*--------------------------------------------------------------------*/
  /* bits: smmc cbbb							*/
  /*--------------------------------------------------------------------*/
  /* s = set/get (0 = get, 1 = set)					*/
  /* m = mode (0 = clear (with s=1), 1 = mode, 2 = byte op, 3 = bit op	*/
  /* c = channel (0 = A, 1 = B, 2 = C)					*/
  /* b = bit (0 to 7)							*/
  /*--------------------------------------------------------------------*/

  spin_lock_irq(&driver_lock);
  switch (cmd & 0xe0) {
     case 0xe0:			/* bit-wise set operation */
        Chan[minor].bits[channel] |= (0x01 << (cmd & 0x07));
	if (dio_has_bit_ops && (channel == 2))
	      /* use fast bit set operator on port C, if available */
	      outb((BYTE)(0x01 | ((cmd & 0x07) << 1)), portbase + CTRL_OFFSET);
	else
              outb((BYTE)(Chan[minor].bits[channel]), portbase + channel);
	break;

     case 0x80:			/* bit-wise clear opearation */
        Chan[minor].bits[channel] &= ~(0x01 << (cmd & 0x07));
	if (dio_has_bit_ops && (channel == 2))
	      /* use fast bit clear operator on port C, if available */
	      outb((BYTE)((cmd & 0x07) << 1), portbase + CTRL_OFFSET);
	else
              outb((BYTE)(Chan[minor].bits[channel]), portbase + channel);
	break;

     case 0xc0:			/* byte-wise set operation */
        Chan[minor].bits[channel] = (BYTE)arg;
        outb((BYTE)arg, portbase + channel);
	break;

     case 0x60:			/* bit-wise get operation */
	put_long(((inb(portbase + channel) & (0x01 << (cmd & 0x07))) > 0)
			? 1L : 0L, (long*)arg);
	break;

     case 0x40:			/* byte-wise get operation */
        put_long(inb(portbase + channel), (long*)arg);
	break;

     case 0xa0:			/* mode-setting operation */
	Chan[minor].mode = (BYTE)arg;
        outb_p((BYTE)(MODE_SET | arg), portbase + CTRL_OFFSET);
	break;

     case 0x20:			/* mode-getting operation */
        put_long(Chan[minor].mode, (long*) arg);
	break;

     case 0x00:			/* no such command */
  	spin_unlock_irq(&driver_lock);
        return(-EINVAL);
	break;
  }
  spin_unlock_irq(&driver_lock);
  return 0;
}

/***************************************************************************
 *
 * See /usr/include/linux/fs.h  for more information
 *
 ***************************************************************************/

/* dictionary-style structure definition */

static struct file_operations dio_fops = {
	owner:		THIS_MODULE,
	read:		dio_read,
	write:		dio_write,
	ioctl:		dio_ioctl,
	open:		dio_open,
	release:	dio_close,
};

/***************************************************************************
 *
 * Loads driver. Called when "insmod dio96.o" is invoked on the command line.
 *
 * NOTICE:  This routine has been written without complete knowledge of
 * the PCI addressing on all the DIO cards.  Operation on PCI_DIO96,
 * PCI_DIO48H, and PCI_DIO24 cards has been confirmed.  For other cards,
 * it may be necessary to change "dio_base_start" appropriately.  However,
 * the proper PCI base address is not necessarily mentioned in the card's
 * manual.
 *
 * Also, some cards have a variant of the 82C55 which does NOT include
 * the bit-wise operators.  DIO port "C" is normally coded for bitwise
 * operations because this method guarantees that bit toggling will not
 * cause signal glitches on other bits in the port.  If a card appears to
 * be responding to commands to ports A and B but not C, try setting
 * "dio_has_bit_ops" to FALSE.
 *
 * Known PCI base addresses (dio_base_start) for specific DIO cards:
 *
 * card		dio_base_start
 * --------------------------------------------------------------
 * PCI-DIO48H:	PCI_BASE_ADDRESS_1
 * PCI-DIO24H:	PCI_BASE_ADDRESS_2
 * PCI-DIO96:	PCI_BASE_ADDRESS_3
 *
 ***************************************************************************/

int init_module(void) 
{
  int i;
  int dio_base_start = PCI_BASE_ADDRESS_3;

  channels = CHANNELS;
  dio_base_reg = BASE_REG;
  dio_has_bit_ops = TRUE;

#ifdef ENABLE_PCI
   {
      char pci_name[12];
      unsigned char pci_irq_line;
	  struct pci_dev *dev = NULL;

      /* Check for up to 8 boards (this should be fixed, but is	*/
      /* unlikely to ever cause anybody grief)			*/

	 /* Need to include clauses for other known board types... */

	 /* Need to clean this up and store everything in a big	   */
	 /* structure and loop through it, rather than this mess   */
	 /* of if/else if statements.				   */

	 /* Currently, I'm only allowing only one board. */

	 if((dev = pci_find_device(PCI_VENDOR_ID_COMPUTERBOARDS,
				 PCI_DEVICE_ID_PCIDIO96, dev))) {
	    strcpy(pci_name, "PCI-DIO96");
	    channels = 6;
	 }
	 else if ((dev = pci_find_device(PCI_VENDOR_ID_COMPUTERBOARDS,
		PCI_DEVICE_ID_PCIDIO96H, dev))) {
	    strcpy(pci_name, "PCI-DIO96H");
	    channels = 6;
	 }
	 else if ((dev = pci_find_device(PCI_VENDOR_ID_COMPUTERBOARDS,
		PCI_DEVICE_ID_PCIDIO48H, dev))) {
	    strcpy(pci_name, "PCI-DIO48H");
	    dio_base_start = PCI_BASE_ADDRESS_1;
	    channels = 4;
	 }
	 else if ((dev = pci_find_device(PCI_VENDOR_ID_COMPUTERBOARDS,
		PCI_DEVICE_ID_PCIDIO24H, dev))) {
	    strcpy(pci_name, "PCI-DIO24H");

            /* AKL - 2003-02-18 - Found PCI-DIO24 needs start base addr 2 */
            dio_base_start = PCI_BASE_ADDRESS_2;
	    channels = 2;
	 }
	 else if ((dev = pci_find_device(PCI_VENDOR_ID_COMPUTERBOARDS,
		PCI_DEVICE_ID_PCIDIO24, dev))) {
	    strcpy(pci_name, "PCI-DIO24");
	    channels = 2;
	 }
	 else if ((dev = pci_find_device(PCI_VENDOR_ID_ICS_ADVENT,
		PCI_DEVICE_ID_PCIDIO24DB, dev))) {
	    strcpy(pci_name, "PCI-DIO24DB");
	    dio_base_start = PCI_BASE_ADDRESS_2;
	    channels = 2;
	 }
	 else if ((dev = pci_find_device(PCI_VENDOR_ID_ICS_ADVENT,
		PCI_DEVICE_ID_PCIDIO24DB_CT, dev))) {
	    strcpy(pci_name, "PCI-DIO24DB/CT");
	    dio_base_start = PCI_BASE_ADDRESS_2;
	    channels = 2;
	 }
	 else if ((dev = pci_find_device(PCI_VENDOR_ID_ICS_ADVENT,
		PCI_DEVICE_ID_PCIDIO24HD, dev))) {
	    strcpy(pci_name, "PCI-DIO24HD");
	    dio_base_start = PCI_BASE_ADDRESS_2;
	    channels = 2;
	 }
	 else if ((dev = pci_find_device(PCI_VENDOR_ID_ICS_ADVENT,
		PCI_DEVICE_ID_PCIDIO24HD_CT, dev))) {
	    strcpy(pci_name, "PCI-DIO24HD/CT");
	    dio_base_start = PCI_BASE_ADDRESS_2;
	    channels = 2;
	 }

	 if(dev) {
		 pci_read_config_byte(dev,PCI_INTERRUPT_LINE,&pci_irq_line);

		 pci_read_config_dword(dev,dio_base_start,&dio_base_reg);

		 dio_base_reg &= PCI_BASE_ADDRESS_IO_MASK;

		 printk(ADAPTER_ID ": PCI BIOS reports ComputerBoards %s "
			"card at I/O %#x, irq %d.\n", pci_name, dio_base_reg,
			pci_irq_line);
	}
   }
#endif
 
  /* Register as a device with kernel.  */

  if (register_chrdev(MajorNumber, "dio96", &dio_fops)) {
     printk(ADAPTER_ID ": Failure to load module.\n");
     return -EIO;
  }

  for (i=0; i < channels; i++)
	request_region(dio_base_reg + i * PORT_OFFSET, 4, "dio96");

  printk(ADAPTER_ID ": address=%#x.", dio_base_reg);
  printk(" driver version 2.4, 04/28/04 tim.edwards@jhuapl.edu\n");

  spin_lock_irq(&driver_lock);

  /* Set all channel structures to show nothing active/open */

  for (i = 0; i < channels; i++)
    Chan[i].open = FALSE;

  spin_unlock_irq(&driver_lock);
  return 0;
}

/***************************************************************************
 *
 * Remove driver. Called when "rmmod dio96" is run on the command line.
 *
 ***************************************************************************/

void cleanup_module(void) 
{
  int i;

  if (unregister_chrdev(MajorNumber, "dio96") != 0) {
    printk(ADAPTER_ID ": cleanup_module failed\n");
  } else {

    for (i=0; i < channels; i++)
       release_region(dio_base_reg + i * PORT_OFFSET, 4);

    printk(ADAPTER_ID ": module removed.\n");

  }    
}

